
// Copyright (c) 1996 Hukatronic (H.C.C.), Tomas Hurka. All Rights Reserved.

#import "DAS_ADC.h"
#import "das_12.h"
#import "DAS_12_public.h"
#import <appkit/nextstd.h>

@implementation DAS_ADC
/*
This class is designed as a user level access to the ADC part of the DAS_12 driver. The class uses locking to reserve the ADC part to itself, so that another instance of the same class cannot mess the hardware registers dealing with ADC.  
*/

extern port_t mig_get_reply_port();

- init
/*
connects to the DAS_12 driver on localhost and initialize DAS_ADC instance. This method also set the default values for PGAGain to 1, scan mode to SINGLE_CHANNEL, start and end channel to 0 , trigger source to SOFTWARE. Returns nil if connection to the DAS_12 driver fails or if ADC part of the driver is locked by another process. If successful returns self.
*/
{
	return [self initOnHost:""];
}

- initOnHost:(const char *)name
/*
connects to the DAS_12 driver on host name and initialize DAS_ADC instance. This method also set the default values for PGAGain to 1, scan mode to SINGLE_CHANNEL, start and end channel to 0 , trigger source to SOFTWARE.  Returns nil if connection to the DAS_12 driver fails or if ADC part of the driver is locked by another process. If successful returns self.
*/
{
	if (![super initOnHost:name])
		return nil;
	if (lockADC(server, mig_get_reply_port())!=KERN_SUCCESS)
	{	NXLogError("%s - ADC locked\n",[self name]);
		[super free];
		return nil;
	}
	[self setPGAGain:1];
	[self setChannels:0:0];
	[self setScanMode:SINGLE_CHANNEL];
	[self setTriggerSource:SOFTWARE];
	return self;
}

- _triggerADC
{
	triggerADC(server);
	return self;
}

- setPGAGain:(unsigned int)gain
/*
Sets PGAGain. The allowed values for gain are 1,2,4,8.
*/
{
	setPGAGain(server,gain);
	return self;
}

- (unsigned int)PGAGain
/*
Returns PGAGain.
*/ 
{	unsigned int g;

	PGAGain(server,&g);
	return g;
}

- setChannels:(int)start:(int)end
/*
Sets start and end channels. If the parameter is NONE, than this parameter is not changed.
*/
{
	setChannels(server,start,end);
	return self;
}

- channels:(int *)start:(int *)end
/*
Returns start and end channel.
*/
{
	channels(server,start,end);
	return self;
}

- setScanMode:(unsigned int)mode
/*
Select scanning mode. The allowed values are SINGLE_CHANNEL, SEQUENTIAL and BURST.
*/
{
	setScanMode(server,mode);
	return self;
}

- (unsigned int)scanMode
/*
Returns actual scan mode.
*/
{	unsigned int m;

	scanMode(server,&m);
	return m;
}
	
- setTriggerSource:(unsigned int)source
/*
Select trigger source. The allowed trigger source is SOFTWARE or EXTERNAL.
*/
{
	setTriggerSource(server,source);
	return self;
}

- (unsigned int)triggerSource
/*
Returns actual trigger source.
*/
{	unsigned int s;

	triggerSource(server,&s);
	return s;
}

- setADCinternalPacingSource:(unsigned)freq
/*
Sets pacing source to internal and also sets the pacing frequency to freq.
*/
{
	setADCinternalPacingSource(server,freq);
	return self;
}

- setADCexternalPacingSource
/*
Sets pacing source to external. The external clock signal must be connected to J3 pin 19.
*/
{
	setADCexternalPacingSource(server);
	return self;
}

- runADC:(unsigned short *)buffer length:(unsigned)len
/*
Starts ADC conversion and reads len 16-bit samples starting at address buffer. If trigger source is SOFTWARE executing this method also triggers the ADC. If trigger source is EXTERNAL call this method first and than trigger the ADC conversion with your external equipment.
*/
{	unsigned loc_len,llen;
	short_ptr in_data;
	
	setDMA(server,len);
	if ([self triggerSource]==SOFTWARE)
		[self _triggerADC];
	for (loc_len=len;loc_len;buffer+=llen,loc_len-=llen)
	{	
		readADC(server,&in_data,&llen);
		if (llen>loc_len)
			llen=loc_len;
		bcopy(in_data,buffer,llen*2);
		vm_deallocate(task_self(), (vm_address_t)in_data,llen*2);
	}
	return self;
}

- runADC:(unsigned short *)buffer length:(unsigned)len frequency:(unsigned)freq
/*
Sets the pacing source to interval with frequency freq and than starts ADC conversion and reads len 16-bit samples starting at address buffer. If trigger source is SOFTWARE executing this method also triggers the ADC. If trigger source is EXTERNAL call this method first and than trigger the ADC with your external equipment.
*/
{
	[self setADCinternalPacingSource:freq];
	return [self runADC:buffer length:len];
}



- free
/*
Unlocks the ADC part of the driver. 
*/
{
	unlockADC(server);
	return [super free];
}

@end