
// Copyright (c) 1996 Hukatronic (H.C.C.), Tomas Hurka. All Rights Reserved.

#import "DAS_DAC.h"
#import "das_12.h"
#import "DAS_12_public.h"
#import <appkit/nextstd.h>

@implementation DAS_DAC
/*
This class is designed as a user level access to the DAC part of the DAS_12 driver. The class uses locking to reserve the DAC part to itself, so that another instance of the same class cannot mess the hardware registers dealing with DAC. There are two different modes for the class. It can be initialized to use external source frequency or to use on-card crystal controlled clock at 1 MHz. If the on-card clock is used, the ADC part of the driver is also locked by DAS_DAC class, since internal clock is common to both ADC and DAC and cannot be shared. On the other hand if the external clock generator is used, both DAC and ADC part can be used simulatenously and only DAC part is locked.
*/

extern port_t mig_get_reply_port();

- init
/*
connects to the DAS_12 driver on localhost and initialize DAS_DAC instance using intnernal on-card 1 MHz clock. Returns nil if connection to the DAS_12 driver fails or if ADC part or DAC part of the driver is locked by another process. If successful returns self.

*/ 
{
	return [self initOnHost:""];
}

- initOnHost:(const char *)name
/*
connects to the DAS_12 driver on host name and initialize DAS_DAC instance using intnernal on-card 1 MHz clock. Returns nil if connection to the DAS_12 driver fails or if ADC part or DAC part of the driver is locked by another process. If successful returns self.
*/ 
{
	if (![super initOnHost:name])
		return nil;
	if (lockDAC(server, mig_get_reply_port())!=KERN_SUCCESS)
	{	NXLogError("%s - DAC locked\n",[self name]);
		[super free];
		return nil;
	}
	if (lockADC(server, mig_get_reply_port())!=KERN_SUCCESS)
	{	NXLogError("%s - ADC locked\n",[self name]);
		[super free];
		return nil;
	}
	setInternalDACSource(server);
	externalSource=NO;
	return self;
}

- initWithExternalDACSource:(unsigned int)freq
/*
connects to the DAS_12 driver on localhost and initialize DAS_DAC instance using external clock. The clock must be connected to J3 pin 21. The frequency of the external clock is specified as freq. This frequency freq is than divided by Timer0 to produce desired frequency for DAC. Returns nil if connection to the DAS_12 driver fails or if DAC part of the driver is locked by another process. If successful returns self.
*/ 
{
	return [self initWithExternalDACSource:freq onHost:""];
}

- initWithExternalDACSource:(unsigned int)freq onHost:(const char *)name
/*
connects to the DAS_12 driver on host name and initialize DAS_DAC instance using external clock. The clock must be connected to J3 pin 21. The frequency of the external clock is specified as freq. This frequency freq is than divided by Timer0 to produce desired frequency for DAC. Returns nil if connection to the DAS_12 driver fails or if DAC part of the driver is locked by another process. If successful returns self.
*/ 
{	
	if (![super initOnHost:name])
		return nil;
	if (lockDAC(server, mig_get_reply_port())!=KERN_SUCCESS)
	{	NXLogError("%s - ADC locked\n",[self name]);
		[super free];
		return nil;
	}
	setExternalDACSource(server,freq);
	externalSource=YES;
	return self;
}

- runDAC:(unsigned int)number buffer:(short *)buf lenght:(unsigned int)len repeat:(unsigned int)rep frequency:(unsigned int)freq
/*
Starts DAC. DAC 0 ro DAC 1 is specified by number. buf points to the beggining of tha 16-bit data samples. The number of the 16-bit samples in  the buffer is len. The whole buffer is send out rep times on frequency freq. Timer0 is used to produce freq from clock frequency (internal or external). The freq cannot be too high. On 486/66 the upper limit is about 6 kHz. On faster machines the upper limit can be higher. The reason is that DAC timing uses interrupt and the handle time of the interrupt routine must me smaller that delay between interrupts. Routine returns imediatelly and the DAC is running on the background. If another DAC is running, the request is not executed and routine returns nil.
*/
{	unsigned offset=(unsigned)buf&(vm_page_size-1);

	if (![self isDACrunning])
	{	runDAC(server,number,buf,len+offset/sizeof(*buf),rep,freq);
		return self;
	}
	return nil;

}

- (BOOL)isDACrunning
/* 
Returns YES if DAC is runnig otherwise returns NO.
*/
{	unsigned int f;
	
	isDACrunning(server,&f);
	return (BOOL)f;
}

- free
/*
Releases all lock acquired by the instance of the class.
*/
{
	if (!externalSource)
		unlockADC(server);
	unlockDAC(server);
	return [super free];
}
@end

