/* 	Copyright (c) 1993 NeXT Computer, Inc.  All rights reserved. 
 *
 * Example of a loadable i386 driver. This driver doesn't do anything
 * other than show that it has been loaded, probed, and has a valid
 * IOEISADeviceDescription.
 *
 * HISTORY
 * 04-Feb-93    Doug Mitchell at NeXT
 *      Created. 
 * 29-Mar-93    Kathy Walrath at NeXT
 *      Modified for use as an example of building, loading, & debugging. 
 */
#import <driverkit/i386/directDevice.h>
#import <driverkit/i386/IOEISADeviceDescription.h>
#import <driverkit/i386/ioPorts.h>
#import <driverkit/generalFuncs.h>
#import <kernserv/kern_server_types.h>
#import <kernserv/prototypes.h>
#import "JoystickDriver.h"
#import "JoystickUser.h"

#define my_PARAMETER_STRING "ack"

static int _JoystickDriverUnit = 0;

@implementation JoystickDriver

+ (IODeviceStyle)deviceStyle
{
    return IO_DirectDevice;
}

/*
 * Probe, configure board and init new instance.  This method is 
 * documented in the IODevice spec sheet.
 */
+ (BOOL)probe:deviceDescription
{
    id driver;
    int js_num;
    int js_count;
    
    js_num = [JoystickDriver find_axes];
    js_count = !!(js_num & 0x3) + !!(js_num & 0xC),

    IOLog("js_init: found %d joystick%c.\n",
	    js_count, (js_num == 1) ? ' ' : 's');

    if (0 == js_count)
    {
	IOLog("No joysticks found.\n");
	return NO;
	/* if the user boots the machine, which runs insmod, and THEN
	    decides to hook up the joystick, well, then we do the wrong
	    thing. But it's a good idea to avoid giving out a false sense
	    of security by letting the module load otherwise. */
    }

    driver = [[self alloc] initFromDeviceDescription:deviceDescription];
    if (driver == nil)
	return NO;
    else return YES;
}

/*
 * Init the new instance.  This method is documented in the i386-specific
 * part of the IODirectDevice spec sheet.
 */
- initFromDeviceDescription:deviceDescription
{
    char name[80];

#ifdef USE_HW
    /* 
     * A real driver would call [super initFromDeviceDescription:], which
     * reserves hardware resources (DMA channels, I/O ports, etc.).  
     * Since this driver doesn't really do anything, it isn't necessary 
     * to reserve resources.  However, it shouldn't hurt anything to 
     * reserve them.  If the resources specified in this driver's bundle 
     * (in /usr/Devices/JoystickDriver.config/*.table) are already reserved,
     * [super initFromDeviceDescription:] will return nil.
     */
    if ([super initFromDeviceDescription:deviceDescription] == nil)
    	return nil;
#endif USE_HW

    IOLog("JoystickDriver: interrupt %d channel %d\n",
	[deviceDescription interrupt], 
	[deviceDescription channel]);
    sprintf(name, "%s%d", my_DEVICE_NAME, _JoystickDriverUnit);
    [self setName:name];
    [self setUnit:_JoystickDriverUnit++];
    [self setLocation:NULL];
    [self registerDevice];
    _myParameter = my_PARAMETER_STRING;
    
    return self; 
}

/* 
 * This method is documented in the IODevice spec sheet.
 */
- (IOReturn)getCharValues   : (unsigned char *)parameterArray
               forParameter : (IOParameterName)parameterName
                      count : (unsigned int *)count
{
    const char  *param;
    unsigned int length;
    unsigned int maxCount = *count;

    if(strcmp(parameterName, my_PARAMETER_NAME) == 0){
        param = _myParameter;
        length = strlen(param);
        if(length >= maxCount) {
            length = maxCount - 1;
        }
        *count = length + 1;
        strncpy(parameterArray, param, length);
        parameterArray[length] = '\0';
        return IO_R_SUCCESS;
    }
    else { 
	/* Pass parameters we don't recognize to our superclass. */
        return [super getCharValues:parameterArray 
            forParameter:parameterName count:count];
    }
}

/* find_axes():

   returns which axes are hooked up, in a bitfield. 2^n is set if
   axis n is hooked up, for 0 <= n < 4.

   REVIEW: should update this to handle eight-axis (four-stick) game port
   cards. anyone have one of these to test on? mattrh 3/23/96 */
+ (int)find_axes
{
    int j;
    static int JS_EXIST;	/* which joysticks' axis exist? */

    outb(0xff, JS_PORT);	/* trigger oneshots */
				/* and see what happens */

    for (j = JS_DEF_TIMEOUT; (0x0f & inb (JS_PORT)) && j; j--)
	    ;	/* do nothing; wait for the timeout */

    JS_EXIST = inb (JS_PORT) & 0x0f; /* get joystick status byte */

    JS_EXIST = (~JS_EXIST) & 0x0f;

    IOLog("find_axes: JS_EXIST is %d (0x%04X)\n", JS_EXIST, JS_EXIST);

    return JS_EXIST;
}

@end
